# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import importlib.metadata

project = "cosmicshear"
copyright = "2022-2026, Nicolas Tessore"
author = "Nicolas Tessore"
version = importlib.metadata.version(project)
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinxcontrib.katex",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



import warnings


def isometry(epsilon, origin):
    """
    An isometry of the ellipticity *epsilon* that maps *origin* to the
    origin.
    """
    xp = epsilon.__array_namespace__()

    return (epsilon - origin) / (1 - xp.conj(origin) * epsilon)


def normal_coordinates(epsilon, origin=None):
    """
    Compute normal coordinates of ellipticity *epsilon* in *origin*
    (default 0).
    """
    xp = epsilon.__array_namespace__()

    if origin is not None:
        epsilon = isometry(epsilon, origin)

    r = xp.abs(epsilon)
    u = epsilon / xp.where(r > 0, r, 1.0)

    return 2 * xp.atanh(r) * u


def exponential_map(vec, origin=None):
    """
    Compute the exponential map of vector *vec* in origin *origin*
    (default 0).
    """
    xp = vec.__array_namespace__()

    r = xp.abs(vec)
    u = vec / xp.where(r > 0, r, 1.0)

    epsilon = xp.tanh(r / 2) * u

    if origin is not None:
        epsilon = isometry(epsilon, -origin)

    return epsilon


def distance(x, y=None):
    """
    Compute the intrinsic distance between ellipticities.
    """
    xp = x.__array_namespace__()

    if y is None:
        return 2 * xp.atanh(xp.abs(x))

    p = xp.abs(x)
    q = xp.abs(y)
    r = xp.abs(x - y)

    return xp.acosh(1 + 2 * r**2 / (((1 - p) * (1 + p)) * ((1 - q) * (1 + q))))


def mean(epsilon, weight=None, *, axis=None, initial=None, maxiter=100, tol=1e-6):
    """
    Compute the Fréchet mean of a sample of ellipticities *epsilon*.  If
    *weight* is given, the mean is weighted.
    """
    xp = epsilon.__array_namespace__()

    if weight is None:
        weight = 1.0

    wtot = xp.sum(0 * epsilon + weight, axis=axis)

    if initial is None:
        initial = xp.sum(weight * epsilon, axis=axis) / wtot

    mu = initial

    # gradient descent rate
    a = 0.5

    for _ in range(maxiter):
        n = normal_coordinates(epsilon, mu)
        v = xp.sum(weight * n, axis=axis) / wtot
        if xp.all(xp.abs(v) < tol):
            break
        mu = exponential_map(a * v, mu)
    else:
        warnings.warn("mean did not converge")

    return mu



__all__ = [
    "chi_from_epsilon",
    "distance",
    "epsilon_from_chi",
    "exponential_map",
    "inverse_transform",
    "isometry",
    "mean",
    "normal_coordinates",
    "transform",
]

from ._ellipticity import (
    chi_from_epsilon,
    epsilon_from_chi,
    transform,
    inverse_transform,
)

from ._manifold import (
    distance,
    exponential_map,
    isometry,
    mean,
    normal_coordinates,
)



def chi_from_epsilon(eps):
    r"""
    Transform epsilon-ellipticity to chi-ellipticity.

    .. math::

        \chi = \frac{2\epsilon}{1 + |\epsilon|^2}

    """
    return (2 * eps) / (1 + eps.real**2 + eps.imag**2)


def epsilon_from_chi(chi):
    r"""
    Transform chi-ellipticity to epsilon-ellipticity.

    .. math::

        \epsilon = \frac{\chi}{1 + \sqrt{1 - |\chi|^2}}

    """
    xp = chi.__array_namespace__()

    return chi / (1 + xp.sqrt(1 - (chi.real**2 + chi.imag**2)))


def transform(jac, eps=None):
    r"""
    Transform epsilon-ellipticity between coordinate systems.

    Uses the Jacobian matrix *jac* (or a stack of matrices).  Transforms
    the ellipticity *eps* if given, or zero ellipticity otherwise.

    Internally, the code computes the shear transformation via the
    decomposition

    .. math::

        \mathrm{J} \begin{pmatrix}
            1 + \chi_1 & \chi_2 \\
            \chi_2 & 1 - \chi_1
        \end{pmatrix} \mathrm{J}^{\mathsf{T}} \\
        = \frac{1}{1 + |\epsilon|^2} \left[
            \mathrm{J} \begin{pmatrix}
                1 + \epsilon_1 & \epsilon_2 \\
                \epsilon_2 & 1 - \epsilon_1
            \end{pmatrix}
        \right]
        \left[
            \mathrm{J} \begin{pmatrix}
                1 + \epsilon_1 & \epsilon_2 \\
                \epsilon_2 & 1 - \epsilon_1
            \end{pmatrix}
        \right]^{\mathsf{T}} \;.

    """
    xp = jac.__array_namespace__()

    if eps is not None:
        a = xp.stack([eps.real, eps.imag, eps.imag, -eps.real], axis=-1)
        jac = jac + (jac @ a.reshape(a.shape[:-1] + (2, 2)))
        del a

    x = (jac[..., 0, 0] + jac[..., 1, 1]) + 1j * (jac[..., 1, 0] - jac[..., 0, 1])
    y = (jac[..., 0, 0] - jac[..., 1, 1]) + 1j * (jac[..., 1, 0] + jac[..., 0, 1])

    # this is y/x.conj() or x/y.conj() without having to deal with the zeros
    return (x * y) / xp.maximum(x.real**2 + x.imag**2, y.real**2 + y.imag**2)


def inverse_transform(jac, eps=None):
    r"""
    Inverse-transform epsilon-ellipticity between coordinate systems.

    Uses the Jacobian matrix *jac* (or a stack of matrices).  Transforms
    the ellipticity *eps* if given, or zero ellipticity otherwise.

    Equivalent to ``transform(inv(jac), eps)``, but does not compute the
    inverse explicitly.

    Internally, the code computes the shear transformation via the
    decomposition

    .. math::

        \mathrm{J}^{-1} \begin{pmatrix}
            1 + \chi_1 & \chi_2 \\
            \chi_2 & 1 - \chi_1
        \end{pmatrix} \mathrm{J}^{-\mathsf{T}} \\
        = \frac{(1 - |\epsilon|^2)^2}{1 + |\epsilon|^2} \left[
            \begin{pmatrix}
                1 - \epsilon_1 & -\epsilon_2 \\
                -\epsilon_2 & 1 + \epsilon_1
            \end{pmatrix} \mathrm{J}
        \right]^{-1}
        \left[
            \begin{pmatrix}
                1 - \epsilon_1 & -\epsilon_2 \\
                -\epsilon_2 & 1 + \epsilon_1
            \end{pmatrix} \mathrm{J}
        \right]^{-\mathsf{T}} \;.

    """
    xp = jac.__array_namespace__()

    if eps is not None:
        a = xp.stack([-eps.real, -eps.imag, -eps.imag, eps.real], axis=-1)
        jac = jac + (a.reshape(a.shape[:-1] + (2, 2)) @ jac)
        del a

    x = (jac[..., 1, 1] + jac[..., 0, 0]) - 1j * (jac[..., 1, 0] - jac[..., 0, 1])
    y = (jac[..., 1, 1] - jac[..., 0, 0]) - 1j * (jac[..., 1, 0] + jac[..., 0, 1])

    # this is y/x.conj() or x/y.conj() without having to deal with the zeros
    return (x * y) / xp.maximum(x.real**2 + x.imag**2, y.real**2 + y.imag**2)



import galsim
import numpy as np
import pytest

import cosmicshear


@pytest.fixture
def random_size():
    return 1000


@pytest.fixture
def eps(random_size, rng):
    a = rng.uniform(0.0, 1.0, random_size)
    t = rng.uniform(-np.pi, np.pi, random_size)
    return a * np.exp(1j * t)


@pytest.fixture
def chi(eps):
    return eps


@pytest.fixture
def jac(random_size, rng):
    from scipy.stats import random_correlation

    eigs = rng.exponential(size=(random_size, 2))
    eigs /= eigs.sum(axis=-1, keepdims=True) / 2

    return np.vectorize(
        random_correlation.rvs,
        otypes=[float],
        excluded={"random_state"},
        signature="(2)->(2, 2)",
    )(eigs, random_state=rng)


def test_chi_from_epsilon(eps):
    chi = cosmicshear.chi_from_epsilon(eps)

    compare = [(s := galsim.Shear(g1=g.real, g2=g.imag)).e1 + 1j * s.e2 for g in eps]

    np.testing.assert_allclose(chi, compare)


def test_epsilon_from_chi(chi):
    eps = cosmicshear.epsilon_from_chi(chi)

    compare = [(s := galsim.Shear(e1=e.real, e2=e.imag)).g1 + 1j * s.g2 for e in chi]

    np.testing.assert_allclose(eps, compare)


def test_transform_zero(jac):
    eps_tfm = cosmicshear.transform(jac)

    compare = [
        galsim.JacobianWCS(*j.ravel()).shearToWorld(galsim.Shear(0j)).shear for j in jac
    ]

    np.testing.assert_allclose(eps_tfm, compare)


def test_transform(jac, eps):
    eps_tfm = cosmicshear.transform(jac, eps)

    compare = [
        galsim.JacobianWCS(*j.ravel()).shearToWorld(galsim.Shear(e)).shear
        for j, e in zip(jac, eps)
    ]

    np.testing.assert_allclose(eps_tfm, compare)


def test_transform_scalar(jac):
    eps_tfm = cosmicshear.transform(jac, 0.1)

    compare = [
        galsim.JacobianWCS(*j.ravel()).shearToWorld(galsim.Shear(0.1 + 0j)).shear
        for j in jac
    ]

    np.testing.assert_allclose(eps_tfm, compare)


def test_inverse_transform_zero(jac):
    eps_tfm = cosmicshear.inverse_transform(jac)

    compare = [
        galsim.JacobianWCS(*j.ravel()).shearToImage(galsim.Shear(0j)).shear for j in jac
    ]

    np.testing.assert_allclose(eps_tfm, compare)


def test_inverse_transform(jac, eps):
    eps_tfm = cosmicshear.inverse_transform(jac, eps)

    compare = [
        galsim.JacobianWCS(*j.ravel()).shearToImage(galsim.Shear(e)).shear
        for j, e in zip(jac, eps)
    ]

    np.testing.assert_allclose(eps_tfm, compare)


def test_transform_roundtrip_zero(jac):
    eps_round = cosmicshear.inverse_transform(jac, cosmicshear.transform(jac))
    np.testing.assert_allclose(eps_round, 0, atol=1e-6, rtol=0.0)


def test_transform_roundtrip(eps, jac):
    eps_round = cosmicshear.inverse_transform(jac, cosmicshear.transform(jac, eps))
    np.testing.assert_allclose(eps_round, eps)
